# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Functions for interpreting and analyzing model output"""

from spyks import core


def state_fun(model, state):
    """Create a lambda that will evaluate functions of the model state"""
    from spyks.codegen import state_replacements
    from spyks.validate import evalf
    repl = state_replacements(model)
    units = [v.units for n, v in model["state"]]
    context = {"X": [state[:, i] * u for i, u in enumerate(units)]}
    context.update(model["parameters"])
    # context = {n: v.magnitude for n, v in model['parameters']}
    # context["X"] = state.T
    return lambda expr: evalf(expr.subs(repl), context)


def currents(model, state):
    """Calculate I_x(t) for a completed model"""
    f = state_fun(model, state)
    return [f(expr).to("pA") for expr in core.currents(model)]


def conductances(model, state):
    """Calculate g_x(t) for a completed model"""
    f = state_fun(model, state)
    return [f(expr).to("nS") for expr in core.conductances(model)]



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Core functions for model descriptors

In spyks, models are specified by a map with following required keys: name
(string), state (mapping), equations (mapping), and parameters (mapping). Optional
keys include forcing (mapping) and reset (with predicate and state).

The convention in this package is that "mapping" is a list of (name, value)
tuples. Values must be specified with physical units (e.g., '5 mV') to permit
dimensional analysis; if units are not specified the quantity is assumed to be
dimensionless. An array, on the other hand, does not have names or quantities.

"""

import os
import sympy as sp
import pint as pq

# need a global unit registry in order for quantities to be compatible
ureg = pq.UnitRegistry()


def n_params(model):
    """ Number of parameters in model """
    return len(model["parameters"])


def n_state(model):
    """ Number of state variables in model """
    return len(model["equations"])


def n_forcing(model):
    """ Number of forcing terms in model """
    return len(model["forcing"])


def param_names(model):
    """ Names of parameters in model """
    return tuple(n for n, v in model["parameters"])


def state_names(model):
    """ Names of state variables in model """
    return tuple(n for n, v in model["state"])


def forcing_names(model):
    """ Names of forcing terms in model """
    return tuple(n for n, v in model["forcing"])


def symbols(model):
    lh = set()
    rh = set()
    for n, eq in model["equations"]:
        lh.add(sp.Symbol(n))
        rh.update(eq.free_symbols)
    try:
        reset = model["reset"]
        rh.update(reset["predicate"].free_symbols)
        for n, eq in reset["state"]:
            rh.update(eq.free_symbols)
        for n, eq in reset.get("clip", []):
            rh.update(eq.free_symbols)
    except KeyError:
        pass
    return lh, rh


def to_array(mapping):
    """ Converts named vector of quantities into a dimensionless array. """
    from numpy import fromiter
    return fromiter((v.magnitude for k, v in mapping), dtype='d', count=len(mapping))


def to_mapping(array, mapping):
    """ Binds values in array to names and units in mapping. Size of array must match mapping. """
    assert len(array) == len(mapping)
    return [(k, x * getattr(v, 'units', v)) for  (k, v), x in zip(mapping, array) ]


def parse_quantity(x):
    if isinstance(x, (float, int)):
        return x * ureg.dimensionless
    else:
        return ureg.parse_expression(x)


def parse(model):
    model['equations'] = [(n, sp.sympify(s)) for n,s in model['equations'].items()]
    try:
        reset = model["reset"]
        reset["predicate"] = sp.sympify(reset["predicate"])
        reset["state"] = [(sp.Symbol(n), sp.sympify(s)) for n,s in reset["state"].items()]
        reset["clip"] = [(sp.Symbol(n), sp.sympify(s)) for n,s in reset["clip"].items()]
    except KeyError:
        pass
    # parse quantities and units
    for k in ('parameters', 'forcing', 'state'):
        if k in model:
            model[k] = [(param, parse_quantity(value)) for param, value in model[k].items()]
    return model


def mapping_updater(new):
    """Returns an f that updates mapping elements with values in new, converting units as needed"""
    def f(kv):
        n, v = kv
        if n in new:
            return (n, parse_quantity(new[n]).to(v.units))
        else:
            return (n, v)
    return f


def load_model(fname):
    """ Loads a model descriptor file. """
    import ruamel.yaml as yaml
    fp = open(fname, "r")
    model = yaml.load(fp, yaml.RoundTripLoader)  # preserves order
    if "base" in model:
        basefile = os.path.join(os.path.dirname(fname), model["base"] + ".yml")
        base = load_model(basefile)              # will get parsed
        if 'parameters' in model:
            base['parameters'] = list(
                map(mapping_updater(dict(model['parameters'])), base['parameters']))
        if 'state' in model:
            base['state'] = list(
                map(mapping_updater(dict(model['state'])), base['state']))
        return base
    else:
        return parse(model)


def update_model(model, **kwargs):
    """Updates model with arrays of values.

    Keyword arguments (e.g., parameters, forcing, state) must match existing
    keys, and array size must match size of model vector.

    """
    for key, array in kwargs.items():
        model[key] = to_mapping(array, model[key])


def currents(model):
    """Extract intrinsic current terms from the model

    This function assumes that the model is biophysical: the model's first
    equation is for voltage, specified by current conservation.

    """
    dV = model["equations"][0][1]
    I_net = sp.fraction(dV)[0]
    forcing_vars = forcing_names(model)
    return (term for term in I_net.args if str(term) not in forcing_vars)


def conductances(model):
    """Extract intrinsic conductance terms from the model

    This function works by dividing out any arguments from currents that depend on the voltage
    """
    V = sp.Symbol("V")
    return (sp.prod(t for t in term.args if V not in t.free_symbols) for term in currents(model))


def conductance_names(model):
    """Extract names of conductances in the model"""
    return (str(expr.args[0]) for expr in currents(model))



# -*- coding: utf-8 -*-
# -*- mode: python -*-

import os
import sys
import sysconfig
import pybind11
import distutils
import logging
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

log = logging.getLogger('spyks')   # root logger


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        return pybind11.get_include(self.user)


def get_boost_include():
    if sys.platform == "win32" :
        return "C:/Boost/include/boost-1_32"
    elif sys.platform == 'darwin':
        return "/opt/local/include"
    else:
        return "/usr/include/boost"


def get_include(*args, **kwargs):
    try:
        from pip import locations
        return os.path.dirname(
            locations.distutils_scheme('spyks', *args, **kwargs)['headers'])
    except ImportError:
        return 'include'


def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class ImportCppExt(Extension):
    """ A custom Extension class for specifying installation destination """
    def __init__(self, libdest, *args, **kwargs):
        self.libdest = libdest
        Extension.__init__(self, *args, **kwargs)


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
            if has_flag(self.compiler, '-ffast-math'):
                opts.append('-ffast-math')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


class CustomDestBuildExt(BuildExt):
    """A custom build extension for specifying target path"""
    def copy_extensions_to_source(self):
        for ext in self.extensions:
            fullname = self.get_ext_fullname(ext.name)
            filename = self.get_ext_filename(fullname)
            src_filename = os.path.join(self.build_lib, filename)
            dest_filename = os.path.join(ext.libdest, os.path.basename(filename))

            distutils.file_util.copy_file(
                src_filename, dest_filename,
                verbose = self.verbose, dry_run = self.dry_run
            )


def should_rebuild(modelfile, cppfile):
    t_model = os.path.getmtime(modelfile)
    t_cpp   = os.path.getmtime(cppfile)
    return t_cpp < t_model


def write_cppfile(model, fname, simplify=True):
    """ Renders model to C++ code and writes to fname"""
    from .codegen import render, simplify_equations
    if simplify:
        log.info("%s: simplifying equations", model["name"])
        model = simplify_equations(model)
    code = render(model)
    with open(fname, "wt") as fp:
        fp.write(code)
    log.info("%s: wrote %s", model["name"], fname)


def build_module(cppfile, name, path, **kwargs):
    import shutil, tempfile
    ext = ImportCppExt(
        path,
        name,
        language = "c++",
        sources = [cppfile],
        include_dirs = [get_pybind_include(), get_pybind_include(True),
                        get_boost_include(), get_include()],
    )

    build_path = tempfile.mkdtemp()
    args = ['build_ext', '--inplace', '-q']
    args.append('--build-temp=' + build_path)
    args.append('--build-lib=' + build_path)
    log.info("%s: compiling extension module", name)
    setup(name = name,
          version = kwargs.get("version", "0.0.1-SNAPSHOT"),
          ext_modules = [ext],
          script_args = args,
          cmdclass = { "build_ext": CustomDestBuildExt })
    shutil.rmtree(build_path)
    log.info("%s: complete - extension module in %s", name, path)


# def import_model(modelfile):
#     from .core import load_model

#     model = load_model(modelfile)
#     log.info("%s: validating model", model["name"])
#     spkv.check_symbols(model)
#     spkv.check_equations(model)
#     path = args.target or os.path.dirname(args.model)
#     cppfile = os.path.join(path, model["name"] + ".cpp")
#     write_cppfile(model, cppfile)

#     path = os.path.dirname(modelfile)
#     cppfile = os.path.join(path, model["name"] +


def compile_script(argv=None):
    from .core import load_model
    import spyks.validate as spkv
    import argparse

    p = argparse.ArgumentParser(description="compile a spyks model file into a python extension module")
    p.add_argument("model", help="the model descriptor file to compile")
    p.add_argument("target", help="the path to put the module (default same as model file)", nargs='?')
    p.add_argument("--skip-compile", help="skip compilation step", action="store_true")
    p.add_argument("--skip-codegen", help="skip code generation step", action="store_true")
    args = p.parse_args(argv)

    ch = logging.StreamHandler()
    formatter = logging.Formatter("[%(name)s] %(message)s")
    loglevel = logging.INFO
    log.setLevel(loglevel)
    ch.setLevel(loglevel)  # change
    ch.setFormatter(formatter)
    log.addHandler(ch)

    model = load_model(args.model)
    log.info("%s: validating model", model["name"])
    spkv.check_symbols(model)
    spkv.check_equations(model)
    path = args.target or os.path.dirname(args.model)
    cppfile = os.path.join(path, model["name"] + ".cpp")
    if not args.skip_codegen:
        write_cppfile(model, cppfile)

    if not args.skip_compile:
        build_module(cppfile, model["name"], path, version=model["version"])



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Generates C++ extension module code for a model

Example:

model = core.load_model(modelfile)
# symbolic simplification is time-intensive but may speed up C++ code
model = simplify_equations(model)
code = render(model)

"""
import logging
import sympy as sp
from spyks.core import n_params, n_state, n_forcing

value_type = "double"
param_var = "p"
state_var = "X"
deriv_var = "dXdt"
forcing_var = "forcing"
time_var = "t"

log = logging.getLogger('spyks')   # root logger


def simplify_equations(model):
    m = model.copy()
    m['equations'] = [(n, expr.simplify()) for n, expr in model['equations']]
    return m


def param_replacements(model):
    """Generate a dictionary to replace named parameters with indexed elements of P"""
    P = sp.IndexedBase(param_var, shape=(n_params(model),))
    return {n: P[i] for i, (n, v) in enumerate(model["parameters"])}


def state_replacements(model):
    """Generate a dictionary to replace named variables with indexed elements of X"""
    X = sp.IndexedBase(state_var, shape=(n_state(model),))
    return {n: X[i] for i, (n, v) in enumerate(model["state"])}


def symbol_replacements(model):
    subs = state_replacements(model)
    subs.update(param_replacements(model))
    return subs


def fmt_subst(name, expr):
    return "const {} {} = {};".format(value_type, sp.ccode(name), sp.ccode(expr))


def fmt_dx(i, expr):
    return "{}[{}] = {};".format(deriv_var, i, sp.ccode(expr))


def fmt_forcing_nd(i, s):
    templ = "const {} {} = {}({}, {});"
    return templ.format(value_type, s, forcing_var, time_var, i)


def fmt_forcing_1d(s):
    templ = "const {} {} = {}({});"
    return templ.format(value_type, s, forcing_var, time_var)


def fmt_reset(n, expr):
    return "{} = {};".format(sp.ccode(n), sp.ccode(expr))


def fmt_clip(n, expr):
    return "if ({0} > {1}) {0} = {1};".format(sp.ccode(n), sp.ccode(expr))


def fmt_systemf(model):
    """Produces c code from expressions, applying cse"""
    repl = symbol_replacements(model)
    if len(model["forcing"]) == 1:
        forcing_s = fmt_forcing_1d(model["forcing"][0][0])
    else:
        forcing_s = "\n".join(fmt_forcing_nd(i, s)
                              for i, (s, v) in enumerate(model["forcing"]))
    log.info("%s: eliminating common subexpressions", model["name"])
    subs, exprs = sp.cse(expr for n, expr in model["equations"])
    subs_s = "\n".join(fmt_subst(n, expr.subs(repl)) for n, expr in subs)
    expr_s = "\n".join(fmt_dx(i, expr.subs(repl)) for i, expr in enumerate(exprs))
    return {
        "forcing": forcing_s,
        "substitutions": subs_s,
        "system": expr_s
    }


def fmt_resetf(model):
    if "reset" not in model: return {}
    repl = symbol_replacements(model)
    out = {
        "reset_predicate": sp.ccode(model['reset']['predicate'].subs(repl)),
        "reset_state": "\n".join(fmt_reset(n.subs(repl), expr.subs(repl)) for n, expr in model['reset']['state'])
    }
    try:
        out["clip"] = "\n".join(fmt_clip(n.subs(repl), expr.subs(repl)) for n, expr in model['reset']['clip'])
    except KeyError:
        pass
    return out


def get_resource(name):
    import posixpath as pp
    import pkg_resources
    return pkg_resources.resource_string("spyks", pp.join("templates", name)).decode("utf-8")


def render(model):
    import string
    if "reset" in model:
        template = "model_reset.cpp"
    else:
        template = "model_continuous.cpp"

    context = dict(name=model["name"],
                   descr=model["description"],
                   head=get_resource("head.cpp"),
                   n_param=n_params(model),
                   n_state=n_state(model),
                   n_forcing=n_forcing(model),
                   param_var=param_var,
                   forcing_var=forcing_var,
                   state_var=state_var,
                   deriv_var=deriv_var,
                   time_var=time_var)
    context.update(fmt_systemf(model))
    context.update(fmt_resetf(model))

    templ = string.Template(get_resource(template))
    return templ.substitute(context)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Generates equations.txt and parameters.txt for makecode """

import logging
from spyks.core import n_params, n_state, n_forcing
from spyks.codegen import simplify_equations

log = logging.getLogger('spyks')   # root logger


def discretize(model):



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- mode: python -*-
import sys
if sys.hexversion < 0x02070000:
    raise RuntimeError("Python 2.7 or higher required")

from spyks import __version__
import os
from setuptools import setup

VERSION = '0.6.1'
cls_txt = """
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
License :: OSI Approved :: GNU General Public License (GPL)
Programming Language :: Python
Topic :: Scientific/Engineering
Topic :: Internet :: WWW/HTTP
Topic :: Internet :: WWW/HTTP :: Dynamic Content
"""

setup(
    name="spyks",
    version=VERSION,
    description="Generate fast c++ code to integrate neuron models",
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.md')).read(),
    classifiers=[x for x in cls_txt.split("\n") if x],

    packages=['spyks'],
    package_data={'spyks': ["templates/*.cpp"]},
    entry_points={'console_scripts': ['spykscc = spyks.build:compile_script'] },
    install_requires = [
        "pybind11>=2.1.1",
        "Pint>=0.7",
        "ruamel.yaml>=0.13",
        "sympy>=1.0"
    ],

    author="Tyler Robbins",
    maintainer='C Daniel Meliza',
)



# -*- coding: utf-8 -*-
# -*- mode: python -*-

from nose.tools import assert_almost_equal, assert_equal, assert_true
import numpy as nx

import spyks as s
import spyks.validate as sv

from models import adex, nakl, biocm

adex_params = [
    {'params': [250, 30, -70.6, 2.0, -55, 144, 4, -70.6, 80.5, 30],
     'forcing': [0.],
     'state': [-70., 0]},
    {'params': (200, 32, -65, 2.0, -56, 120, 5, -70.6, 80.5, 30),
     'forcing': [0.05],
     'state': [-50., 0]},
    {'params': nx.asarray([200, 32, -65, 2.0, -56, 120, 5, -70.6, 80.5, 30]),
     'forcing': [0.05],
     'state': [50., 10]},
]


def test_adex_dxdt():
    pymodel = s.load_model("models/adex.yml")
    def compare_adex(params, forcing, state):
        # inj = models.timeseries(forcing, 0.05)
        model = adex.model(params, forcing, 0.05)
        dXdt = model(state, 0)
        s.update_model(pymodel, state=state, forcing=forcing, parameters=params)
        pydXdt = s.to_array(sv.deriv(pymodel))
        assert_true(nx.allclose(dXdt, pydXdt))
    for tvals in adex_params:
        yield compare_adex, tvals['params'], tvals['forcing'], tvals['state']


def test_adex_reset():
    pymodel = s.load_model("models/adex.yml")
    def compare_reset(params, forcing, state):
        model = adex.model(params, forcing, 0.05)
        reset, new_state = model.reset(state)
        s.update_model(pymodel, state=state, forcing=forcing, parameters=params)
        pypost_state = sv.reset(pymodel)
        if reset:
            pypost_state = s.to_array(pypost_state)
            assert_true(nx.allclose(new_state, pypost_state))
        else:
            assert_equal(reset, pypost_state)
            assert_true(nx.allclose(state, new_state))
    for tvals in adex_params:
        yield compare_reset, tvals['params'], tvals['forcing'], tvals['state']


def test_adex_integration():
    I = 500
    N = 1000
    dt = 0.05
    params = adex_params[0]['params']
    x0 = adex_params[0]['state']
    data = nx.ones(N) * I
    model = adex.model(params, data, dt)
    X = adex.integrate(params, x0, data, dt, dt)
    X2 = adex.integrate(model, x0, data.size * dt, dt)
    assert_true(nx.allclose(X, X2))
    # with these parameters, there should be exactly one spike at 555
    events = (X[:,0] > 29.9).nonzero()[0]
    assert_equal(events.size, 1)
    assert_equal(events[0], 555)
    return X


nakl_params = [
    {'params': [1., 120., 50., 20., -77., 0.3, -54.4,
                -40., 15., 0.1, 0.4, -40., 15., -60., -15.,
                1., 7., -60., -15., -55., 30., 1., 5., -55., -30.],
     'forcing': [0.],
     'state': [-70., 0., 0., 0.]},
    {'params': [1., 120., 50., 20., -77., 0.3, -54.4,
                -40., 15., 0.1, 0.4, -40., 15., -60., -15.,
                1., 7., -60., -15., -55., 30., 1., 5., -55., -30.],
     'forcing': [100.],
     'state': [-70., 0., 0.2, 0.1]},
    {'params': [1.02, 70.7, 55., .38, -85., 0.054, -65,
                -40., 15., 0.1, 0.4, -40., 15., -60., -15.,
                1., 7., -60., -15., -55., 30., 1., 5., -55., -30.],
     'forcing': [0.],
     'state': [-65., 0.1, 0.2, 0.1]},
]

def test_nakl_dxdt():
    pymodel = s.load_model("models/nakl.yml")
    def compare_nakl(params, forcing, state):
        model = nakl.model(params, forcing, 0.05)
        dXdt = model(state, 0)
        s.update_model(pymodel, state=state, forcing=forcing, parameters=params)
        pydXdt = s.to_array(sv.deriv(pymodel))
        assert_true(nx.allclose(dXdt, pydXdt))
    for tvals in nakl_params:
        yield compare_nakl, tvals['params'], tvals['forcing'], tvals['state']


def test_nakl_integration(I=50):
    N = 10000
    dt = 0.05
    params = nakl_params[0]['params']
    x0 = nakl_params[0]['state']
    data = nx.ones(N) * I
    model = nakl.model(params, data, dt)
    X = nakl.integrate(params, x0, data, dt, dt)
    X2 = nakl.integrate(model, x0, data.size * dt, dt)
    assert_true(nx.allclose(X, X2))
    return X


def test_biocm():
    pymodel = s.load_model("models/biocm.yml")
    params = s.to_array(pymodel['parameters'])
    forcing = s.to_array(pymodel['forcing'])
    state = s.to_array(pymodel['state'])
    model = biocm.model(params, forcing, 0.05)
    dXdt = model(state.tolist(), 0)
    pydXdt = s.to_array(sv.deriv(pymodel))
    assert_true(nx.allclose(dXdt, pydXdt))


def test_biocm_integration(I=20):
    N = 5000
    dt = 0.05
    pymodel = s.load_model("models/biocm.yml")
    params = s.to_array(pymodel['parameters'])
    x0 = s.to_array(pymodel['state'])
    data = nx.ones(N) * I
    model = biocm.model(params, data, dt)
    X = biocm.integrate(params, x0, data, dt, dt)
    X2 = biocm.integrate(model, x0, data.size * dt, dt)
    assert_true(nx.allclose(X, X2))
    return X
